#include "grtest.h"
#include <getopt.h>

#define GRSEC_OPT_NOCOLOR 100
#define KERNEL_CONFIG_FILE "/boot/config"

char config_grkernsec[100][100];

char *kernel_config;

int grsec_test_nocolor = 0;
int grsec_test_verbose = 0;
int grsec_test_failed = 0;

pid_t grtest_pid;

struct regression_func
{
    int (*func)();
    const char *name;
    const char *desc;
	const char *config;
	const char *sysctl;
};

int chroot_caps();
int chroot_chdir();
int chroot_chmod();
int chroot_chroot();
int chroot_fchdir();
int chroot_kill();
int chroot_mknod();
int chroot_nice();
int chroot_nonroot_shm();
int chroot_ptrace();
int chroot_shm();
int chroot_sysctl();
int chroot_unix();
int chroot_mount();
int grsec_locked();
int chroot_pivot();
int execve_limiting();
int fifo();
int memkmemport();
int rawio();
int gr_symlink();
int gr_dmesg();
int gr_link();
int tpe();
int proc_user();

static struct regression_func myfuncs[] = {
	// chroot tests
	{
		&chroot_chdir,
		"chroot_chdir",
		"enforce chdir(\"/\") on a chroot call",
		"CONFIG_GRKERNSEC_CHROOT_CHDIR",
		"chroot_enforce_chdir"
	},
	{
		&chroot_caps,
		"chroot_caps",
		"deny dangerous capabilities in chroot",
		"CONFIG_GRKERNSEC_CHROOT_CAPS",
		"chroot_caps"
	},
	{
		&chroot_chmod,
		"chroot_chmod",
		"deny fchmod +s and chmod +s in chroot",
		"CONFIG_GRKERNSEC_CHROOT_CHMOD",
		"chroot_deny_chmod"
	},
	{
		&chroot_chroot,
		"chroot_chroot",
		"deny double chroot outside of chroot",
		"CONFIG_GRKERNSEC_CHROOT_DOUBLE",
		"chroot_deny_chroot"
	},
	{
		&chroot_fchdir,
		"chroot_fchdir",
		"deny fchdir out of chroot",
		"CONFIG_GRKERNSEC_CHROOT_FCHDIR",
		"chroot_deny_fchdir"
	},
	{
		&chroot_pivot,
		"chroot_pivot",
		"deny pivot_root in chroot",
		"CONFIG_GRKERNSEC_CHROOT_PIVOT",
		"chroot_deny_pivot"
	},
	{
		&chroot_mknod,
		"chroot_mknod",
		"deny mknod in chroot",
		"CONFIG_GRKERNSEC_CHROOT_MKNOD",
		"chroot_deny_mknod"
	},
	{
		&chroot_sysctl,
		"chroot_sysctl",
		"deny sysctl in chroot",
		"CONFIG_GRKERNSEC_CHROOT_SYSCTL",
		"chroot_deny_sysctl"
	},
	{
		&chroot_unix,
		"chroot_unix",
		"deny connect to abstract unix domain socket out of chroot",
		"CONFIG_GRKERNSEC_CHROOT_UNIX",
		"chroot_deny_unix"
	},
	{
		&chroot_nice,
		"chroot_nice",
		"deny nice and priority raise in chroot",
		"CONFIG_GRKERNSEC_CHROOT_NICE",
		"chroot_restrict_nice"
	},
	{
		&chroot_nonroot_shm,
		"chroot_nonroot_shma",
		"deny shared memory attach out of chroot (as non-root)",
		"CONFIG_GRKERNSEC_CHROOT_SHMAT", // is this right??
		"chroot_deny_shmat"
	},
	{
		&chroot_shm,
		"chroot_shm",
		"deny shared memory attach out of chroot",
		"CONFIG_GRKERNSEC_CHROOT_SHMAT",
		"chroot_deny_shmat"
	},
	// TODO: combind kill and ptrace, they are both affected by chroot_findtask
	{
		&chroot_kill,
		"chroot_kill",
		"kill out of chroot",
		"CONFIG_GRKERNSEC_CHROOT_FINDTASK",
		"chroot_findtask"
	},
	{
		&chroot_ptrace,
		"chroot_ptrace",
		"deny ptrace out of chroot",
		"CONFIG_GRKERNSEC_CHROOT_FINDTASK",
		"chroot_findtask"
	},
	{
		&chroot_mount,
		"chroot_mount",
		"deny mounts in a chroot",
		"CONFIG_GRKERNSEC_CHROOT_MOUNT",
		"chroot_deny_mount"
	},
	// non-chroot tests
	{
		&tpe,
		"tpe",
		"trusted path execution",
		"CONFIG_GRKERNSEC_TPE",
		"tpe"
	},
	{
		&gr_dmesg,
		"dmesg",
		"dmesg restriction",
		"CONFIG_GRKERNSEC_DMESG",
		"dmesg"
	},
	{
		&fifo,
		"fifo_restrictions",
		"fifo restrictions",
		"CONFIG_GRKERNSEC_FIFO",
		"fifo_restrictions"
	},
	{
		&gr_link,
		"linking_restrictions",
		"hardlink restrictions",
		"CONFIG_GRKERNSEC_LINK",
		"linking_restrictions"
	},
	{
		&gr_symlink,
		"symlink",
		"symlink restrictions",
		"CONFIG_GRKERNSEC_LINK", // is this right??
		"linking_restrictions"
	},
	{
		&execve_limiting,
		"execve_limiting",
		"execve resource limiting",
		"CONFIG_GRKERNSEC_EXECVE",
		"execve_limiting"
	},
   	{
		&grsec_locked,
		"grsec_locked",
		"grsecurity sysctl lock",
		"CONFIG_GRKERNSEC_SYSCTL",
		"grsec_lock"
	},
	{
		&proc_user,
		"proc_user",
		"proc user restrictions",
		"CONFIG_GRKERNSEC_PROC_USER",
		NULL
	},
	{
		&memkmemport,
		"memkmemport",
		"deny write/mmap/open of /dev/mem, /dev/port, and /dev/kmem",
		"CONFIG_GRKERNSEC_KMEM",
		NULL
	},
	{
		&rawio,
		"rawio",
		"deny ioperm and iopl",
		"CONFIG_GRKERNSEC_IO",
		NULL
	},
	{ NULL, NULL, NULL, NULL, NULL }
};

void colorized_print(char *msg, char *color) {
	if (!grsec_test_nocolor)
		printf("%s", color);
	printf("%s", msg);
	if (!grsec_test_nocolor)
		printf("\e[0m");
}

int grtest_error(char *msg) {
	
	colorized_print("ERROR", "\e[33;1m");
	
	printf(" %s", msg);
	if ( getpid() == grtest_pid ) {
		return GRSEC_TEST_ERROR;
	}
	exit(GRSEC_TEST_ERROR);
}

int grsec_sysctl_read(char *name) {
	int rc = 0;
	char *tmp;
	char buf[1024];
	FILE *fp;

	if (!name || !*name) {
		return -1;
	}

	// allocate memory for the sysctl name... path + file + null byte
	tmp = malloc(strlen(GRSEC_PROC_PATH)+strlen(name)+1);
	// build full path into *tmp
	strcpy(tmp, GRSEC_PROC_PATH);
	strcat(tmp, name);

	fp = fopen(tmp, "r");

	if (!fp) {
		rc = -1;
	} else {
		if(fgets(buf, sizeof buf, fp)) {
			// strip off newline char
			char *nl = strchr(buf,'\n');
			if (nl) *nl = '\0';
			// convert first character of string to integer
			rc = (int)buf[0]-48;
			if (rc < 0)
				rc = -1;
		} else {
			rc = -1;
		}
		fclose(fp);
	}

	return rc;
}

void grsec_config_read() {
	char *lineptr, *tmp;
	int i = 0;
	size_t n;
    FILE *fp;
    struct utsname buf;

	if(kernel_config) {
		fp = fopen(kernel_config, "r");

		if (!fp) {
			perror("Unable to open kernel config file");
			exit(1);
		}
	} else {

		fp = fopen(KERNEL_CONFIG_FILE, "r");
		
		if (!fp) {
			
			uname(&buf);

			tmp = malloc(strlen(KERNEL_CONFIG_FILE)+strlen("-")+strlen(buf.release)+1);
			strcpy(tmp, KERNEL_CONFIG_FILE);
			strcat(tmp, "-");
			strcat(tmp, buf.release);
			
			fp = fopen(tmp, "r");
			
		}
		
	}

    if (fp) {

		while( getline(&lineptr, &n, fp) != -1 ) {

			if(lineptr[0] == '#') continue;
			if(lineptr[0] == '\n') continue;
			if(strncmp(lineptr, "CONFIG_GRKERNSEC", strlen("CONFIG_GRKERNSEC"))) continue;
			
			// strip out the = character and everything behind it
            char *ln = strchr(lineptr,'=');
            if (ln) *ln = '\0';
			
			strcpy(config_grkernsec[i], lineptr);
			i++;

		}

		config_grkernsec[i][0] = '\0';

        fclose(fp);
    } else {
		perror("Unable to open kernel config file");
		exit(1);
	}

}

int grsec_config_find(char *config) {
	int i;

	if(config == NULL) return 0;

	for(i = 0; i < 100; i++) {
		if(!strcmp( config_grkernsec[i], config)) {
			return 1;
		}
	}
	
	return 0;
}

void do_grsec_test(struct regression_func myfunc)
{
	int status, ret;
	pid_t pid;

	grtest_pid = getpid();
	
	if (myfunc.desc)
		printf("%-60s ", myfunc.desc);
	else if (myfunc.name)
		printf("%-60s ", myfunc.name);
	else
		printf("%-60s ", "Undefined function call");
	
	fflush(stdout);
	
	pid = fork();

	if ( pid == 0 ) {
		
		status = (*myfunc.func)();
		exit(status);
		
	} else if (pid > 0) {
		wait(&status);
		status = status / 256;
	} else if (pid < 0) {
		perror("fork");
		return;
	}

	switch (status) {
		
	case GRSEC_TEST_PASSED:
		colorized_print("PASSED", "\e[32;1m");
		break;
    case GRSEC_TEST_FAILED:

        ret = grsec_sysctl_read(myfunc.sysctl);

        if (ret == 0) // sysctl found, but turned off
            colorized_print("DISABLED", "\e[33;1m");
        else if (ret == 1) // found and turned on, genuinely failed
            colorized_print("FAILED", "\e[31;1m");
        else if (grsec_config_find(myfunc.config)) // genuinely failed
			colorized_print("FAILED", "\e[31;1m");
		else // test not compiled into the kernel
			colorized_print("N/A", "\e[33;1m");

        break;
	case GRSEC_TEST_NA:
		colorized_print("N/A", "\e[33;1m");
		break;
	case GRSEC_TEST_FORKFAIL:
		printf("Fork failed!");
		break;
	default:
		break;
	}

    if (status) grsec_test_failed++;
	
	printf("\n");	
}

void usage()
{
	int i;

	printf("Usage: grtest [OPTIONS]\n"
		   "\t-c\t--config\tManually specify the path to the .config file your kernel was built with\n"
		   "\t-q\t--quiet\t\tBe quiet (only print failed tests)\n"
		   "\t--nocolor\t\tDon't display ASCII color codes\n"
		   "\t-h\t--help\t\tShow this help\n"
		   "\t-a\t--all\t\tRun all tests\n"
		   "\t-t\t--test\t\tRun a specific test\n\n"
		   "\tTests will result in one of the following:\n\n"
		   "\t* PASSED - the feature works as expected\n"
		   "\t* FAILED - the feature doesn't work as expected\n"
		   "\t* DISABLED - the test failed, but the sysctl option for this feature is turned off\n"
		   "\t* N/A - the test failed, but the kernel config option for this feature doesn't appear to be enabled\n\n"
		   "\tThe exit status of grtest will be the number of tests that didn't pass\n\n"
		   "\tImplemented tests (used with the -t option):\n\n");

	for ( i = 0; myfuncs[i].name != NULL; i++ ) {
		printf("\t%s\n", myfuncs[i].name);
	}
	
	exit(1);
}

int main(int argc, char *argv[])
{
	int i;
	int next_option = 0;
	int doall = 0;
	int status = 0;

	struct regression_func func = { 0 };
	
	pid_t pid;
	uid_t uid;
	
	const char *const short_opts = "at:c:hv";
	
	static struct option long_opts[] = {
		{"all", 0, NULL, 'a'},
		{"config", 1, NULL, 'c'},
		{"help", 0, NULL, 'h'},
		{"nocolor", 0, NULL, GRSEC_OPT_NOCOLOR},
		{"test", 1, NULL, 't'},
		{"verbose", 0, NULL, 'v'},
		{NULL, 0, NULL, 0},
	};
	
	uid = getuid();
	
	if (uid) {
		printf("Tests must be run as root.\n");
		exit(1);
	}
	
	if (argc < 2) {
		printf("not enough args\n");
		usage();
	}

	while ((next_option = getopt_long(argc, argv, short_opts, long_opts, NULL)) != -1) {
		
		switch (next_option) {
			
		case 'a':
			doall = 1;
			break;
			
		case 'c':

			kernel_config = malloc(strlen(optarg));
			strcpy(kernel_config, optarg);

			break;

		case 'h':
			usage();
			
			break;
			
		case GRSEC_OPT_NOCOLOR:
			
			grsec_test_nocolor = 1;
			
			break;
			
		case 't':
			for ( i = 0; myfuncs[i].name != NULL; i++ ) {
				if (!strcmp(optarg, myfuncs[i].name))
					func = myfuncs[i];
			}
			
			if (func.name == NULL) {
				printf("Function unimplemented: %s\n", optarg);
				usage();
			}
			
			break;
			
		case 'v':
			
			grsec_test_verbose++;
			
			break;
			
		default:
			usage();
			break;
		}
		
	}

	grsec_config_read();

	if (doall == 1) {
		
		for ( i = 0; myfuncs[i].name != NULL; i++ ) {
			do_grsec_test(myfuncs[i]);
		}

	} else if (func.name != NULL) {
		do_grsec_test(func);
	}

	return grsec_test_failed;
}

